#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Start the PyQt graphical interface to gaitutils

@author: Jussi (jnu@iki.fi)
"""

from gaitutils.gui import _gaitmenu

if __name__ == '__main__':
    _gaitmenu.main()
